"""Class implementation for the polygon-related classes' x2 mix-in.
"""

from apysc._display.polygon_apply_current_points_mixin import (
    PolygonApplyCurrentPointsMixIn,
)
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.int import Int


class PolygonX2MixIn(PolygonApplyCurrentPointsMixIn):

    _x2: Int

    @property
    @add_debug_info_setting(module_name=__name__)
    def x2(self) -> Int:
        """
        Get a second x-coordinate.

        Returns
        -------
        x2 : Int
            A second x-coordinate.
        """
        x2: Int = self._points[1].x
        return x2

    @x2.setter
    @add_debug_info_setting(module_name=__name__)
    def x2(self, value: Int) -> None:
        import apysc as ap

        self._x2._value = value._value
        new_point: ap.Point2D = self._points[1]
        new_point.x = value
        self._points[1] = new_point
        self._apply_current_points()



"""Class implementation for the `Polygon`-related classes'
`_append_constructor_expression` mix-in.
"""

from typing_extensions import final

from apysc._display.points_2d_mixin import Points2DMixIn
from apysc._display.points_var_name_mixin import PointsVarNameMixIn
from apysc._html.debug_mode import add_debug_info_setting


class PolygonAppendConstructorExpressionMixIn(Points2DMixIn, PointsVarNameMixIn):
    @final
    @add_debug_info_setting(module_name=__name__)
    def _append_constructor_expression(self) -> None:
        """
        Append a polygon's constructor expression.
        """
        import apysc as ap
        from apysc._display.graphics_base import GraphicsBase

        stage: ap.Stage = ap.get_stage()
        points_var_name: str
        points_expression: str
        points_var_name, points_expression = self._make_2dim_points_expression()
        expression: str = (
            f"{points_expression}"
            f"\nvar {self.variable_name} = {stage.variable_name}"
            f"\n  .polygon({points_var_name})"
            "\n  .attr({"
        )
        if isinstance(self, GraphicsBase):
            expression = self._append_basic_vals_expression(
                expression=expression, indent_num=2
            )
        expression += "\n  });"
        ap.append_js_expression(expression=expression)
        self._points_var_name = points_var_name



"""Class implementation for the polygon-related classes' y1 mix-in.
"""

from apysc._display.polygon_apply_current_points_mixin import (
    PolygonApplyCurrentPointsMixIn,
)
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.int import Int


class PolygonY1MixIn(PolygonApplyCurrentPointsMixIn):

    _y1: Int

    @property
    @add_debug_info_setting(module_name=__name__)
    def y1(self) -> Int:
        """
        Get a first y-coordinate.

        Returns
        -------
        y1 : Int
            A first y-coordinate.
        """
        y1: Int = self._points[0].y
        return y1

    @y1.setter
    @add_debug_info_setting(module_name=__name__)
    def y1(self, value: Int) -> None:
        """
        Set a first y-coordinate.

        Parameters
        ----------
        value : Int
            Y-coordinate to set.
        """
        import apysc as ap

        self._y1._value = value._value
        new_point: ap.Point2D = self._points[0]
        new_point.y = value
        self._points[0] = new_point
        self._apply_current_points()



"""This module is for the translation mapping data of the
following document:

Document file: path_bezier_3d.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {
    ##################################################
    "# PathBezier3D class": "# PathBezier3D クラス",
    ##################################################
    "This page explains the `PathBezier3D` class.": "このページでは`PathBezier3D`クラスについて説明します。",  # noqa
    ##################################################
    "## What class is this?": "## クラス概要",
    ##################################################
    "The `PathBezier3D` class is the class to set a 3D bezier curve on a path.": "`PathBezier3D`クラスはパス上に3次のベジェ曲線を設定するためのクラスです。",  # noqa
    ##################################################
    "This class has two control points (as a comparison, the 2D bezier class has one control point).": "このクラスは2つの制御点を持ちます（比較対象として、2次のベジェ曲線では1つの制御点のみを持ちます）。",  # noqa
    ##################################################
    "Mainly, the `Path` class constructor or `draw_path` interfaces use this setting.": "主にこの設定は`Path`クラスのコンストラクタと`draw_path`メソッドのインターフェイスで使用されます。",  # noqa
    ##################################################
    "## Basic usage": "## 基本的な使い方",
    ##################################################
    "The `PathBezier3D` class constructor requires the `control_x1`, `control_y1`, `control_x2`, `control_y2`, `dest_x`, and `dest_y` arguments.": "`PathBezier3D`クラスのコンストラクタでは`control_x1`、`control_y1`、`control_x2`、`control_y2`、`dest_x`、`dest_y`の各引数の指定が必要です。",  # noqa
    ##################################################
    "The `control_x1` and `control_y1` are the coordinates to determine the first control point of the bezier curve.": "`control_x1`と`control_y1`はベジェ曲線の最初の制御点の位置を決定するための座標指定です。",  # noqa
    ##################################################
    "Similarly, the `control_x2` and `control_y2` are the coordinates to determine the second control point of the bezier curve.": "同様に、`control_x2`と`control_y2`はベジェ曲線の2番目の制御点の位置を決定するための座標指定です。",  # noqa
    ##################################################
    "The `dest_x` and `dest_y` are the bezier curve's destination coordinates.": "`dest_x`と`dest_y`はベジェ曲線の終点座標の指定となります。",  # noqa
    ##################################################
    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color="#333", stage_width=250, stage_height=270, stage_elem_id="stage"\n)\n\npath: ap.Path = ap.Path(\n    path_data_list=[\n        ap.PathMoveTo(x=50, y=200),\n        ap.PathBezier3D(\n            control_x1=50,\n            control_y1=25,\n            control_x2=200,\n            control_y2=25,\n            dest_x=200,\n            dest_y=200,\n        ),\n    ],\n    line_color="#0af",\n    line_thickness=5,\n)\n\nap.save_overall_html(dest_dir_path="path_bezier_3d_basic_usage_1/")\n```': '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color="#333", stage_width=250, stage_height=270, stage_elem_id="stage"\n)\n\npath: ap.Path = ap.Path(\n    path_data_list=[\n        ap.PathMoveTo(x=50, y=200),\n        ap.PathBezier3D(\n            control_x1=50,\n            control_y1=25,\n            control_x2=200,\n            control_y2=25,\n            dest_x=200,\n            dest_y=200,\n        ),\n    ],\n    line_color="#0af",\n    line_thickness=5,\n)\n\nap.save_overall_html(dest_dir_path="path_bezier_3d_basic_usage_1/")\n```',  # noqa
    ##################################################
    "In the following example, the cyan circle shows the first control point of the bezier curve.": "以下のコード例では、シアン色の円はベジェ曲線の最初の制御点を示しています。",  # noqa
    ##################################################
    "The magenta circle shows the second control point of the bezier curve.": "マゼンタの円ではベジェ曲線の2つ目の制御点を示しています。",  # noqa
    ##################################################
    "And also, the yellow circle shows the destination point of the bezier curve.": "また、黄色の円ではベジェ曲線の終点座標を示しています。",  # noqa
    ##################################################
    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color="#333", stage_width=250, stage_height=270, stage_elem_id="stage"\n)\n\nCONTROL_X1: int = 50\nCONTROL_Y1: int = 25\nCONTROL_X2: int = 200\nCONTROL_Y2: int = 25\nDEST_X: int = 200\nDEST_Y: int = 200\n\npath: ap.Path = ap.Path(\n    path_data_list=[\n        ap.PathMoveTo(x=50, y=200),\n        ap.PathBezier3D(\n            control_x1=CONTROL_X1,\n            control_y1=CONTROL_Y1,\n            control_x2=CONTROL_X2,\n            control_y2=CONTROL_Y2,\n            dest_x=DEST_X,\n            dest_y=DEST_Y,\n        ),\n    ],\n    line_color="#fff",\n    line_thickness=5,\n)\n\nRADIUS: int = 10\ncyan_circle: ap.Circle = ap.Circle(\n    x=CONTROL_X1,\n    y=CONTROL_Y1,\n    radius=RADIUS,\n    fill_color="#0af",\n)\n\nmagenta_circle: ap.Circle = ap.Circle(\n    x=CONTROL_X2,\n    y=CONTROL_Y2,\n    radius=RADIUS,\n    fill_color="#f0a",\n)\n\nyellow_circle: ap.Circle = ap.Circle(\n    x=DEST_X,\n    y=DEST_Y,\n    radius=RADIUS,\n    fill_color="#ff0",\n)\n\nap.save_overall_html(dest_dir_path="path_bezier_3d_basic_usage_2/")\n```': '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color="#333", stage_width=250, stage_height=270, stage_elem_id="stage"\n)\n\nCONTROL_X1: int = 50\nCONTROL_Y1: int = 25\nCONTROL_X2: int = 200\nCONTROL_Y2: int = 25\nDEST_X: int = 200\nDEST_Y: int = 200\n\npath: ap.Path = ap.Path(\n    path_data_list=[\n        ap.PathMoveTo(x=50, y=200),\n        ap.PathBezier3D(\n            control_x1=CONTROL_X1,\n            control_y1=CONTROL_Y1,\n            control_x2=CONTROL_X2,\n            control_y2=CONTROL_Y2,\n            dest_x=DEST_X,\n            dest_y=DEST_Y,\n        ),\n    ],\n    line_color="#fff",\n    line_thickness=5,\n)\n\nRADIUS: int = 10\ncyan_circle: ap.Circle = ap.Circle(\n    x=CONTROL_X1,\n    y=CONTROL_Y1,\n    radius=RADIUS,\n    fill_color="#0af",\n)\n\nmagenta_circle: ap.Circle = ap.Circle(\n    x=CONTROL_X2,\n    y=CONTROL_Y2,\n    radius=RADIUS,\n    fill_color="#f0a",\n)\n\nyellow_circle: ap.Circle = ap.Circle(\n    x=DEST_X,\n    y=DEST_Y,\n    radius=RADIUS,\n    fill_color="#ff0",\n)\n\nap.save_overall_html(dest_dir_path="path_bezier_3d_basic_usage_2/")\n```',  # noqa
    ##################################################
    "## Relative position setting": "## 相対座標設定",
    ##################################################
    "The constructor's `relative` optional argument changes its behavior.": "コンストラクタの`relative`のオプション引数はその挙動を変更します。",  # noqa
    ##################################################
    "For example, if you set True to its argument, coordinates become relative.": "例として、もしその引数にTrueを指定した場合座標は相対座標として設定されます。",  # noqa
    ##################################################
    "The default setting is False, and it becomes absolute.": "デフォルト値はFalseとなっており、この設定では絶対座標として扱われます。",  # noqa
    ##################################################
    "A criteria point is a starting point, neither a first control point nor a second control point.": "基準点は開始位置の座標となります。最初の制御点や2つ目の制御点ではありません。",  # noqa
    ##################################################
    "The following example sets the relative setting and draws the bezier curve.": "以下のコード例ではrelativeの設定をしてベジェ曲線を描画しています。",  # noqa
    ##################################################
    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color="#333", stage_width=250, stage_height=270, stage_elem_id="stage"\n)\n\npath: ap.Path = ap.Path(\n    path_data_list=[\n        ap.PathMoveTo(x=50, y=200),\n        ap.PathBezier3D(\n            control_x1=0,\n            control_y1=-175,\n            control_x2=150,\n            control_y2=-175,\n            dest_x=150,\n            dest_y=0,\n            relative=True,\n        ),\n    ],\n    line_color="#0af",\n    line_thickness=5,\n)\n\nap.save_overall_html(dest_dir_path="path_bezier_3d_relative/")\n```': '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color="#333", stage_width=250, stage_height=270, stage_elem_id="stage"\n)\n\npath: ap.Path = ap.Path(\n    path_data_list=[\n        ap.PathMoveTo(x=50, y=200),\n        ap.PathBezier3D(\n            control_x1=0,\n            control_y1=-175,\n            control_x2=150,\n            control_y2=-175,\n            dest_x=150,\n            dest_y=0,\n            relative=True,\n        ),\n    ],\n    line_color="#0af",\n    line_thickness=5,\n)\n\nap.save_overall_html(dest_dir_path="path_bezier_3d_relative/")\n```',  # noqa
    ##################################################
    "## PathBezier3D class constructor API": "## PathBezier3D クラスのコンストラクタのAPI",
    ##################################################
    '<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>': '<span class="inconspicuous-txt">特記事項: このAPIドキュメントはドキュメントビルド用のスクリプトによって自動で生成・同期されています。そのためもしかしたらこの節の内容は前節までの内容と重複している場合があります。</span>',  # noqa
    ##################################################
    "**[Interface summary]**": "**[インターフェイス概要]**",
    ##################################################
    "Path data class for the SVG's `3D bezier curve` (C).<hr>": "SVGの3次のベジェ曲線（C）のパスデータのためのクラスです。<hr>",  # noqa
    ##################################################
    "**[Parameters]**": "**[引数]**",
    ##################################################
    "- `control_x1`: Int or int": "- `control_x1`: Int or int",
    ##################################################
    "  - X-coordinate of the bezier's first control point.": "  - ベジェ曲線の最初の制御点のX座標。",
    ##################################################
    "- `control_y1`: Int or int": "- `control_y1`: Int or int",
    ##################################################
    "  - Y-coordinate of the bezier's first control point.": "  - ベジェ曲線の最初の制御点のY座標。",
    ##################################################
    "- `control_x2`: Int or int": "- `control_x2`: Int or int",
    ##################################################
    "  - X-coordinate of the bezier's second control point.": "  - ベジェ曲線の2つ目の制御点のX座標。",  # noqa
    ##################################################
    "- `control_y2`: Int or int": "- `control_y2`: Int or int",
    ##################################################
    "  - Y-coordinate of the bezier's second control point.": "  - ベジェ曲線の2つ目の制御点のY座標。",  # noqa
    ##################################################
    "- `dest_x`: Int or int": "- `dest_x`: Int or int",
    ##################################################
    "  - X-coordinate of the destination point.": "  - 終点のX座標。",
    ##################################################
    "- `dest_y`: Int or int": "- `dest_y`: Int or int",
    ##################################################
    "  - Y-coordinate of the destination point.": "  - 終点のY座標。",
    ##################################################
    "- `relative`: bool or Boolean, default False": "- `relative`: bool or Boolean, default False",  # noqa
    ##################################################
    "  - A boolean value indicates whether the path coordinates are relative or not (absolute).": "  - パスの座標が相対座標として扱うかもしくは絶対座標として扱うかどうかの真偽値。",  # noqa
    ##################################################
    "- `variable_name_suffix`: str, default ''": "- `variable_name_suffix`: str, default ''",  # noqa
    ##################################################
    "  - A JavaScript variable name suffix string. This setting is sometimes useful for JavaScript debugging.": "  - JavaScript上の変数のサフィックスの設定です。この設定はJavaScriptのデバッグ時に役立つことがあります。",  # noqa
    ##################################################
    "<hr>": "<hr>",
    ##################################################
    "**[Examples]**": "**[コードサンプル]**",
    ##################################################
    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.line_style(color="#fff", thickness=3)\n>>> path: ap.Path = sprite.graphics.draw_path(\n...     path_data_list=[\n...         ap.PathMoveTo(x=0, y=50),\n...         ap.PathBezier3D(\n...             control_x1=0,\n...             control_y1=0,\n...             control_x2=50,\n...             control_y2=0,\n...             dest_x=50,\n...             dest_y=50,\n...         ),\n...         ap.PathBezier3DContinual(\n...             control_x=100, control_y=100, dest_x=100, dest_y=50\n...         ),\n...     ]\n... )\n```': '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.line_style(color="#fff", thickness=3)\n>>> path: ap.Path = sprite.graphics.draw_path(\n...     path_data_list=[\n...         ap.PathMoveTo(x=0, y=50),\n...         ap.PathBezier3D(\n...             control_x1=0,\n...             control_y1=0,\n...             control_x2=50,\n...             control_y2=0,\n...             dest_x=50,\n...             dest_y=50,\n...         ),\n...         ap.PathBezier3DContinual(\n...             control_x=100, control_y=100, dest_x=100, dest_y=50\n...         ),\n...     ]\n... )\n```',  # noqa
    ##################################################
    "<hr>": "<hr>",
    ##################################################
    "**[References]**": "**[関連資料]**",
    ##################################################
    "- [Path class](https://simon-ritchie.github.io/apysc/en/path.html)": "- [Path クラス](https://simon-ritchie.github.io/apysc/jp/jp_path.html)",  # noqa
    ##################################################
    "- [Graphics draw_path interface](https://simon-ritchie.github.io/apysc/en/graphics_draw_path.html)": "- [Graphics クラスの draw_path インターフェイス](https://simon-ritchie.github.io/apysc/jp/jp_graphics_draw_path.html)",  # noqa
    ##################################################
    "- [PathBezier3DContinual class](https://simon-ritchie.github.io/apysc/en/path_bezier_3d_continual.html)": "- [PathBezier3DContinual クラス](https://simon-ritchie.github.io/apysc/jp/jp_path_bezier_3d_continual.html)",  # noqa
}



"""Class implementation for the polygon-related classes' y3 mix-in.
"""

from apysc._display.polygon_apply_current_points_mixin import (
    PolygonApplyCurrentPointsMixIn,
)
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.int import Int


class PolygonY3MixIn(PolygonApplyCurrentPointsMixIn):

    _y3: Int

    @property
    @add_debug_info_setting(module_name=__name__)
    def y3(self) -> Int:
        """
        Get a third y-coordinate.

        Returns
        -------
        y3 : Int
            A third y-coordinate.
        """
        y3: Int = self._points[2].y
        return y3

    @y3.setter
    def y3(self, value: Int) -> None:
        """
        Set a third y-coordinate.

        Parameters
        ----------
        value : Int
            Y-coordinate to set.
        """
        import apysc as ap

        self._y3._value = value._value
        new_point: ap.Point2D = self._points[2]
        new_point.y = value
        self._points[2] = new_point
        self._apply_current_points()



"""Implementations of Triangle class.
"""

from typing import Optional
from typing import Union

from typing_extensions import final

from apysc._display import graphics
from apysc._display.child_mixin import ChildMixIn
from apysc._display.fill_alpha_mixin import FillAlphaMixIn
from apysc._display.fill_color_mixin import FillColorMixIn
from apysc._display.graphics_base import GraphicsBase
from apysc._display.line_caps import LineCaps
from apysc._display.line_dash_dot_setting import LineDashDotSetting
from apysc._display.line_dash_setting import LineDashSetting
from apysc._display.line_dot_setting import LineDotSetting
from apysc._display.line_joints import LineJoints
from apysc._display.line_round_dot_setting import LineRoundDotSetting
from apysc._display.polygon_append_constructor_expression_mixin import (
    PolygonAppendConstructorExpressionMixIn,
)
from apysc._display.polygon_x1_mixin import PolygonX1MixIn
from apysc._display.polygon_x2_mixin import PolygonX2MixIn
from apysc._display.polygon_x3_mixin import PolygonX3MixIn
from apysc._display.polygon_y1_mixin import PolygonY1MixIn
from apysc._display.polygon_y2_mixin import PolygonY2MixIn
from apysc._display.polygon_y3_mixin import PolygonY3MixIn
from apysc._display.x_mixin import XMixIn
from apysc._display.y_mixin import YMixIn
from apysc._geom.point2d import Point2D
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.array import Array
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.string import String
from apysc._type.variable_name_suffix_mixin import VariableNameSuffixMixIn
from apysc._validation import arg_validation_decos


class Triangle(
    XMixIn,
    YMixIn,
    PolygonX1MixIn,
    PolygonY1MixIn,
    PolygonX2MixIn,
    PolygonY2MixIn,
    PolygonX3MixIn,
    PolygonY3MixIn,
    GraphicsBase,
    FillColorMixIn,
    FillAlphaMixIn,
    PolygonAppendConstructorExpressionMixIn,
    VariableNameSuffixMixIn,
):
    """
    The triangle vector graphics class.

    References
    ----------
    - Triangle class
        - https://simon-ritchie.github.io/apysc/en/triangle.html

    Examples
    --------
    >>> import apysc as ap
    >>> _ = ap.Stage()
    >>> triangle: ap.Triangle = ap.Triangle(
    ...     x1=75,
    ...     y1=50,
    ...     x2=50,
    ...     y2=100,
    ...     x3=100,
    ...     y3=100,
    ...     fill_color="#0af",
    ...     line_color="#fff",
    ...     line_thickness=3,
    ... )
    >>> triangle.x2
    Int(50)
    >>> triangle.y1 = ap.Int(30)
    >>> triangle.y1
    Int(30)
    """

    # self
    @arg_validation_decos.multiple_line_settings_are_not_set(arg_position_index=0)
    # x1
    @arg_validation_decos.is_integer(arg_position_index=1)
    # y1
    @arg_validation_decos.is_integer(arg_position_index=2)
    # x2
    @arg_validation_decos.is_integer(arg_position_index=3)
    # y2
    @arg_validation_decos.is_integer(arg_position_index=4)
    # x3
    @arg_validation_decos.is_integer(arg_position_index=5)
    # y3
    @arg_validation_decos.is_integer(arg_position_index=6)
    # fill_color
    @arg_validation_decos.is_hex_color_code_format(arg_position_index=7)
    # fill_alpha
    @arg_validation_decos.num_is_0_to_1_range(arg_position_index=8)
    # line_color
    @arg_validation_decos.is_hex_color_code_format(arg_position_index=9)
    # line_alpha
    @arg_validation_decos.num_is_0_to_1_range(arg_position_index=10)
    # line_thickness
    @arg_validation_decos.is_integer(arg_position_index=11)
    @arg_validation_decos.num_is_gte_zero(arg_position_index=11)
    # line_cap
    @arg_validation_decos.is_line_cap(arg_position_index=12, optional=True)
    # line_joints
    @arg_validation_decos.is_line_joints(arg_position_index=13, optional=True)
    # line_dot_setting
    @arg_validation_decos.is_line_dot_setting(arg_position_index=14)
    # line_dash_setting
    @arg_validation_decos.is_line_dash_setting(arg_position_index=15)
    # line_round_dot_setting
    @arg_validation_decos.is_line_round_dot_setting(arg_position_index=16)
    # line_dash_dot_setting
    @arg_validation_decos.is_line_dash_dot_setting(arg_position_index=17)
    # parent
    @arg_validation_decos.is_display_object_container(
        arg_position_index=18, optional=True
    )
    # variable_name_suffix
    @arg_validation_decos.is_builtin_string(arg_position_index=19, optional=False)
    @add_debug_info_setting(module_name=__name__)
    def __init__(
        self,
        *,
        x1: Union[int, Int],
        y1: Union[int, Int],
        x2: Union[int, Int],
        y2: Union[int, Int],
        x3: Union[int, Int],
        y3: Union[int, Int],
        fill_color: Union[str, String] = "",
        fill_alpha: Union[float, Number] = 1.0,
        line_color: Union[str, String] = "",
        line_alpha: Union[float, Number] = 1.0,
        line_thickness: Union[int, Int] = 1,
        line_cap: Optional[Union[String, LineCaps]] = None,
        line_joints: Optional[Union[String, LineJoints]] = None,
        line_dot_setting: Optional[LineDotSetting] = None,
        line_dash_setting: Optional[LineDashSetting] = None,
        line_round_dot_setting: Optional[LineRoundDotSetting] = None,
        line_dash_dot_setting: Optional[LineDashDotSetting] = None,
        parent: Optional[ChildMixIn] = None,
        variable_name_suffix: str = "",
    ) -> None:
        """
        Create a triangle vector graphics instance.

        Parameters
        ----------
        x1 : Union[int, Int]
            First vertex's x coordinate.
        y1 : Union[int, Int]
            First vertex's y coordinate.
        x2 : Union[int, Int]
            Second vertex's x coordinate.
        y2 : Union[int, Int]
            Second vertex's y coordinate.
        x3 : Union[int, Int]
            Third vertex's x coordinate.
        y3 : Union[int, Int]
            Third vertex's y coordinate.
        fill_color : str or String, default ''
            A fill-color to set.
        fill_alpha : float or Number, default 1.0
            A fill-alpha to set.
        line_color : str or String, default ''
            A line-color to set.
        line_alpha : float or Number, default 1.0
            A line-alpha to set.
        line_thickness : int or Int, default 1
            A line-thickness (line-width) to set.
        line_cap : String or LineCaps or None, default None
            A line-cap setting to set.
        line_joints : String or LineJoints or None, default None
            A line-joints setting to set.
        line_dot_setting : LineDotSetting or None, default None
            A dot setting to set.
        line_dash_setting : LineDashSetting or None, default None
            A dash setting to set.
        line_round_dot_setting : LineRoundDotSetting or None, default None
            A round-dot setting to set.
        line_dash_dot_setting : LineDashDotSetting or None, default None
            A dash dot (1-dot chain) setting to set.
        parent : ChildMixIn or None, default None
            A parent instance to add this instance.
            If a specified value is None, this interface uses
            a stage instance.
        variable_name_suffix : str, default ''
            A JavaScript variable name suffix string.
            This setting is sometimes useful for JavaScript debugging.

        References
        ----------
        - Triangle class
            - https://simon-ritchie.github.io/apysc/en/triangle.html

        Examples
        --------
        >>> import apysc as ap
        >>> _ = ap.Stage()
        >>> triangle: ap.Triangle = ap.Triangle(
        ...     x1=75,
        ...     y1=50,
        ...     x2=50,
        ...     y2=100,
        ...     x3=100,
        ...     y3=100,
        ...     fill_color="#0af",
        ...     line_color="#fff",
        ...     line_thickness=3,
        ... )
        >>> triangle.x2
        Int(50)
        >>> triangle.y1 = ap.Int(30)
        >>> triangle.y1
        Int(30)
        """
        from apysc._converter.to_apysc_val_from_builtin import (
            get_copied_int_from_builtin_val,
        )
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names

        self._variable_name_suffix = variable_name_suffix
        variable_name: str = expression_variables_util.get_next_variable_name(
            type_name=var_names.TRIANGLE,
        )
        self.variable_name = variable_name
        self._set_initial_basic_values(
            fill_color=fill_color,
            fill_alpha=fill_alpha,
            line_color=line_color,
            line_thickness=line_thickness,
            line_alpha=line_alpha,
            line_cap=line_cap,
            line_joints=line_joints,
        )
        self._x1 = get_copied_int_from_builtin_val(integer=x1)
        self._y1 = get_copied_int_from_builtin_val(integer=y1)
        self._x2 = get_copied_int_from_builtin_val(integer=x2)
        self._y2 = get_copied_int_from_builtin_val(integer=y2)
        self._x3 = get_copied_int_from_builtin_val(integer=x3)
        self._y3 = get_copied_int_from_builtin_val(integer=y3)
        self._set_points_with_each_coordinate()
        self._append_constructor_expression()
        self._set_line_setting_if_not_none_value_exists(
            line_dot_setting=line_dot_setting,
            line_dash_setting=line_dash_setting,
            line_round_dot_setting=line_round_dot_setting,
            line_dash_dot_setting=line_dash_dot_setting,
        )
        self._set_x_and_y_with_minimum_point()
        super(Triangle, self).__init__(parent=parent, variable_name=variable_name)

    @final
    @add_debug_info_setting(module_name=__name__)
    def _set_points_with_each_coordinate(self) -> None:
        """
        Set the `_points`' attribute value with each coordinate.
        """
        self._points = Array(
            value=[
                Point2D(x=self._x1, y=self._y1),
                Point2D(x=self._x2, y=self._y2),
                Point2D(x=self._x3, y=self._y3),
            ]
        )

    @final
    def _set_x_and_y_with_minimum_point(self) -> None:
        """
        Set an x and y properties coordinate with a minimum point.
        """
        min_x: int = min(self._x1._value, self._x2._value, self._x3._value)
        min_y: int = min(self._y1._value, self._y2._value, self._y3._value)

        suffix: str = self._get_attr_or_variable_name_suffix(value_identifier="x")
        self._x = Int(min_x, variable_name_suffix=suffix)

        suffix = self._get_attr_or_variable_name_suffix(value_identifier="y")
        self._y = Int(min_y, variable_name_suffix=suffix)

    @classmethod
    @final
    def _create_with_graphics(
        cls,
        *,
        graphics: "graphics.Graphics",
        x1: Union[int, Int],
        y1: Union[int, Int],
        x2: Union[int, Int],
        y2: Union[int, Int],
        x3: Union[int, Int],
        y3: Union[int, Int],
        variable_name_suffix: str = "",
    ) -> "Triangle":
        """
        Create a triangle instance with the instance of
        specified graphics.

        Parameters
        ----------
        graphics : graphics.Graphics
            Graphics instance to link this instance.
        x1 : Union[int, Int]
            First vertex's y coordinate.
        y1 : Union[int, Int]
            First vertex's y coordinate.
        x2 : Union[int, Int]
            Second vertex's x coordinate.
        y2 : Union[int, Int]
            Second vertex's y coordinate.
        x3 : Union[int, Int]
            Third vertex's x coordinate.
        y3 : Union[int, Int]
            Third vertex's y coordinate.
        variable_name_suffix : str, optional
            A JavaScript variable name suffix string.
            This setting is sometimes useful for JavaScript debugging.

        Returns
        -------
        triangle : Triangle
            A created triangle instance.
        """
        triangle: Triangle = Triangle(
            x1=x1,
            y1=y1,
            x2=x2,
            y2=y2,
            x3=x3,
            y3=y3,
            fill_color=graphics._fill_color,
            fill_alpha=graphics._fill_alpha,
            line_color=graphics._line_color,
            line_alpha=graphics._line_alpha,
            line_thickness=graphics._line_thickness,
            line_cap=graphics._line_cap,
            line_joints=graphics._line_joints,
            line_dot_setting=graphics._line_dot_setting,
            line_dash_setting=graphics._line_dash_setting,
            line_round_dot_setting=graphics._line_round_dot_setting,
            line_dash_dot_setting=graphics._line_dash_dot_setting,
            parent=graphics,
            variable_name_suffix=variable_name_suffix,
        )
        return triangle

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Triangle('<variable_name>')`).
        """
        repr_str: str = f"Triangle('{self.variable_name}')"
        return repr_str



"""Class implementation for the animation_fill_color mix-in.
"""

from typing import TypeVar
from typing import Union

from typing_extensions import final

from apysc._animation.animation_fill_color import AnimationFillColor
from apysc._animation.animation_mixins import AnimationMixIns
from apysc._animation.easing import Easing
from apysc._type.int import Int
from apysc._type.string import String
from apysc._validation import arg_validation_decos

StrOrString = TypeVar("StrOrString", str, String)


class AnimationFillColorMixIn(AnimationMixIns):
    @final
    @arg_validation_decos.is_hex_color_code_format(arg_position_index=1)
    @arg_validation_decos.is_integer(arg_position_index=2)
    @arg_validation_decos.num_is_gt_zero(arg_position_index=2)
    @arg_validation_decos.is_integer(arg_position_index=3)
    @arg_validation_decos.is_easing(arg_position_index=4)
    def animation_fill_color(
        self,
        *,
        fill_color: StrOrString,
        duration: Union[int, Int] = 3000,
        delay: Union[int, Int] = 0,
        easing: Easing = Easing.LINEAR
    ) -> AnimationFillColor:
        """
        Set the fill color animation setting.

        Notes
        -----
        To start this animation, you need to call the `start` method of
        the returned instance.

        Parameters
        ----------
        fill_color : str or String
            The final fill color (hex color code) of the animation.
        duration : Int or int, default 3000
            Milliseconds before an animation ends.
        delay : Int or int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Returns
        -------
        animation_fill_color : AnimationFillColor
            Created animation setting instance.

        References
        ----------
        - animation_fill_color interface
            - https://simon-ritchie.github.io/apysc/en/animation_fill_color.html
        - Animation interfaces duration setting
            - https://simon-ritchie.github.io/apysc/en/animation_duration.html
        - Animation interfaces delay setting
            - https://simon-ritchie.github.io/apysc/en/animation_delay.html
        - Each animation interface return value
            - https://simon-ritchie.github.io/apysc/en/animation_return_value.html  # noqa
        - Sequential animation setting
            - https://simon-ritchie.github.io/apysc/en/sequential_animation.html
        - animation_parallel interface
            - https://simon-ritchie.github.io/apysc/en/animation_parallel.html
        - Easing enum
            - https://simon-ritchie.github.io/apysc/en/easing_enum.html

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color="#0af")
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50
        ... )
        >>> _ = rectangle.animation_fill_color(
        ...     fill_color="#f0a",
        ...     duration=1500,
        ...     easing=ap.Easing.EASE_OUT_QUINT,
        ... ).start()
        """
        animation_fill_color: AnimationFillColor = AnimationFillColor(
            target=self,
            fill_color=fill_color,
            duration=duration,
            delay=delay,
            easing=easing,
        )
        return animation_fill_color



"""Class implementation for the animation_fill_alpha interface.
"""

from typing import Union

from typing_extensions import final

from apysc._animation.animation_fill_alpha import AnimationFillAlpha
from apysc._animation.animation_mixins import AnimationMixIns
from apysc._animation.easing import Easing
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._validation import arg_validation_decos


class AnimationFillAlphaMixIn(AnimationMixIns):
    @final
    @arg_validation_decos.is_num(arg_position_index=1)
    @arg_validation_decos.num_is_0_to_1_range(arg_position_index=1)
    @arg_validation_decos.is_integer(arg_position_index=2)
    @arg_validation_decos.num_is_gt_zero(arg_position_index=2)
    @arg_validation_decos.is_integer(arg_position_index=3)
    @arg_validation_decos.is_easing(arg_position_index=4)
    def animation_fill_alpha(
        self,
        *,
        alpha: Union[float, Number],
        duration: Union[int, Int] = 3000,
        delay: Union[int, Int] = 0,
        easing: Easing = Easing.LINEAR
    ) -> AnimationFillAlpha:
        """
        Set the fill alpha (opacity) animation setting.

        Notes
        -----
        To start this animation, you need to call the `start` method of
        the returned instance.

        Parameters
        ----------
        alpha : Number or float
            The final alpha (opacity) of the animation.
        duration : Int or int, default 3000
            Milliseconds before an animation ends.
        delay : Int or int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Returns
        -------
        animation_fill_alpha : AnimationFillAlpha
            Created animation setting instance.

        References
        ----------
        - animation_fill_alpha interface
            - https://simon-ritchie.github.io/apysc/en/animation_fill_alpha.html
        - Animation interfaces duration setting
            - https://simon-ritchie.github.io/apysc/en/animation_duration.html
        - Each animation interface return value
            - https://simon-ritchie.github.io/apysc/en/animation_return_value.html  # noqa
        - Sequential animation setting
            - https://simon-ritchie.github.io/apysc/en/sequential_animation.html
        - animation_parallel interface
            - https://simon-ritchie.github.io/apysc/en/animation_parallel.html
        - Easing enum
            - https://simon-ritchie.github.io/apysc/en/easing_enum.html

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color="#0af")
        >>> circle: ap.Circle = sprite.graphics.draw_circle(x=100, y=100, radius=50)
        >>> _ = circle.animation_y(
        ...     y=100,
        ...     duration=1500,
        ...     easing=ap.Easing.EASE_OUT_QUINT,
        ... ).start()
        """
        animation_fill_alpha: AnimationFillAlpha = AnimationFillAlpha(
            target=self, alpha=alpha, duration=duration, delay=delay, easing=easing
        )
        return animation_fill_alpha



"""Class implementations for the hour-related mix-in.
"""

from typing import Union

from typing_extensions import final

from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.int import Int
from apysc._type.revert_mixin import RevertMixIn
from apysc._type.variable_name_mixin import VariableNameMixIn
from apysc._type.variable_name_suffix_attr_or_var_mixin import (
    VariableNameSuffixAttrOrVarMixIn,
)
from apysc._validation import arg_validation_decos


class HourMixIn(VariableNameMixIn, VariableNameSuffixAttrOrVarMixIn, RevertMixIn):

    _initial_hour: Union[int, Int]
    _hour: Int

    @final
    @add_debug_info_setting(module_name=__name__)
    @arg_validation_decos.is_hour_int(arg_position_index=1)
    def _set_init_hour_value(self, *, hour: Union[int, Int]) -> None:
        """
        Set an initial hour value.

        Parameters
        ----------
        hour : Union[int, Int]
            An hour value to set.
        """
        from apysc._converter.to_apysc_val_from_builtin import (
            get_copied_int_from_builtin_val,
        )

        self._initial_hour = hour
        suffix: str = self._get_attr_or_variable_name_suffix(value_identifier="hour")
        self._hour = get_copied_int_from_builtin_val(
            integer=hour, variable_name_suffix=suffix
        )

    @final
    @add_debug_info_setting(module_name=__name__)
    def _get_init_hour_argument_expression(self) -> str:
        """
        Get an initial hour's argument expression string.

        Returns
        -------
        expression : str
            A created expression string.
        """
        if isinstance(self._initial_hour, Int):
            return f", {self._initial_hour.variable_name}"
        return f", {self._hour._value}"

    @property
    @add_debug_info_setting(module_name=__name__)
    def hour(self) -> Int:
        """
        Get a current hour's value.

        Returns
        -------
        hour : Int
            A current hour value.

        References
        ----------
        - DateTime class hour property
            - https://simon-ritchie.github.io/apysc/en/datetime_hour.html

        Examples
        --------
        >>> import apysc as ap
        >>> datetime_: ap.DateTime = ap.DateTime(year=2022, month=12, day=1, hour=5)
        >>> datetime_.hour
        Int(5)
        >>> datetime_.hour = ap.Int(10)
        >>> datetime_.hour
        Int(10)
        """
        copied_hour_val: Int = self._hour._copy()
        self._append_hour_getter_expression(hour_val=copied_hour_val)
        return copied_hour_val

    @hour.setter
    @arg_validation_decos.is_hour_int(arg_position_index=1)
    @add_debug_info_setting(module_name=__name__)
    def hour(self, value: Int) -> None:
        """
        Set an hour value.

        Parameters
        ----------
        value : Int
            A hour value to set.

        References
        ----------
        - DateTime class hour property
            - https://simon-ritchie.github.io/apysc/en/datetime_hour.html

        Examples
        --------
        >>> import apysc as ap
        >>> datetime_: ap.DateTime = ap.DateTime(year=2022, month=12, day=1, hour=5)
        >>> datetime_.hour
        Int(5)
        >>> datetime_.hour = ap.Int(10)
        >>> datetime_.hour
        Int(10)
        """
        self._hour = value._copy()
        self._append_hour_setter_expression(hour_val=value)

    @final
    @arg_validation_decos.is_hour_int(arg_position_index=1)
    @add_debug_info_setting(module_name=__name__)
    def _append_hour_getter_expression(self, *, hour_val: Int) -> None:
        """
        Append an hour's getter expression string.

        Parameters
        ----------
        hour_val : Int
            An hour value to use in an expression.
        """
        import apysc as ap

        expression: str = f"{hour_val.variable_name} = {self.variable_name}.getHours();"
        ap.append_js_expression(expression=expression)

    @final
    @arg_validation_decos.is_hour_int(arg_position_index=1)
    @add_debug_info_setting(module_name=__name__)
    def _append_hour_setter_expression(self, *, hour_val: Int) -> None:
        """
        Append an hour's setter expression string.

        Parameters
        ----------
        hour_val : Int
            An hour value to use in an expression.
        """
        import apysc as ap

        expression: str = f"{self.variable_name}.setHours({hour_val.variable_name});"
        ap.append_js_expression(expression=expression)

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._hour._run_all_make_snapshot_methods(snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if a snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._hour._run_all_revert_methods(snapshot_name=snapshot_name)



"""Implementations of the Rectangle class.
"""

from typing import Optional
from typing import Union

from typing_extensions import final

from apysc._display import graphics
from apysc._display.child_mixin import ChildMixIn
from apysc._display.ellipse_height_mixin import EllipseHeightMixIn
from apysc._display.ellipse_width_mixin import EllipseWidthMixIn
from apysc._display.fill_alpha_mixin import FillAlphaMixIn
from apysc._display.fill_color_mixin import FillColorMixIn
from apysc._display.graphics_base import GraphicsBase
from apysc._display.height_mixin import HeightMixIn
from apysc._display.line_caps import LineCaps
from apysc._display.line_dash_dot_setting import LineDashDotSetting
from apysc._display.line_dash_setting import LineDashSetting
from apysc._display.line_dot_setting import LineDotSetting
from apysc._display.line_joints import LineJoints
from apysc._display.line_round_dot_setting import LineRoundDotSetting
from apysc._display.width_mixin import WidthMixIn
from apysc._display.x_mixin import XMixIn
from apysc._display.y_mixin import YMixIn
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.string import String
from apysc._type.variable_name_suffix_mixin import VariableNameSuffixMixIn
from apysc._validation import arg_validation_decos


class Rectangle(
    XMixIn,
    YMixIn,
    GraphicsBase,
    WidthMixIn,
    HeightMixIn,
    EllipseWidthMixIn,
    EllipseHeightMixIn,
    FillColorMixIn,
    FillAlphaMixIn,
    VariableNameSuffixMixIn,
):
    """
    The rectangle vector graphics class.

    References
    ----------
    - Rectangle class
        - https://simon-ritchie.github.io/apysc/en/rectangle.html
    - Graphics draw_rect interface
        - https://simon-ritchie.github.io/apysc/en/graphics_draw_rect.html

    Examples
    --------
    >>> import apysc as ap
    >>> stage: ap.Stage = ap.Stage()
    >>> sprite: ap.Sprite = ap.Sprite()
    >>> sprite.graphics.begin_fill(color="#0af")
    >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
    ...     x=50, y=50, width=100, height=75
    ... )
    >>> rectangle.x
    Int(50)

    >>> rectangle.y
    Int(50)

    >>> rectangle.width
    Int(100)

    >>> rectangle.height
    Int(75)

    >>> rectangle.fill_color
    String('#00aaff')
    """

    # self
    @arg_validation_decos.multiple_line_settings_are_not_set(arg_position_index=0)
    # x
    @arg_validation_decos.is_integer(arg_position_index=1)
    # y
    @arg_validation_decos.is_integer(arg_position_index=2)
    # width
    @arg_validation_decos.is_integer(arg_position_index=3)
    @arg_validation_decos.num_is_gte_zero(arg_position_index=3)
    # height
    @arg_validation_decos.is_integer(arg_position_index=4)
    @arg_validation_decos.num_is_gte_zero(arg_position_index=4)
    # ellipse_width
    @arg_validation_decos.is_integer(arg_position_index=5)
    @arg_validation_decos.num_is_gte_zero(arg_position_index=5)
    # ellipse_height
    @arg_validation_decos.is_integer(arg_position_index=6)
    @arg_validation_decos.num_is_gte_zero(arg_position_index=6)
    # fill_color
    @arg_validation_decos.is_hex_color_code_format(arg_position_index=7)
    # fill_alpha
    @arg_validation_decos.num_is_0_to_1_range(arg_position_index=8)
    # line_color
    @arg_validation_decos.is_hex_color_code_format(arg_position_index=9)
    # line_alpha
    @arg_validation_decos.num_is_0_to_1_range(arg_position_index=10)
    # line_thickness
    @arg_validation_decos.is_integer(arg_position_index=11)
    @arg_validation_decos.num_is_gte_zero(arg_position_index=11)
    # line_cap
    @arg_validation_decos.is_line_cap(arg_position_index=12, optional=True)
    # line_joints
    @arg_validation_decos.is_line_joints(arg_position_index=13, optional=True)
    # line_dot_setting
    @arg_validation_decos.is_line_dot_setting(arg_position_index=14)
    # line_dash_setting
    @arg_validation_decos.is_line_dash_setting(arg_position_index=15)
    # line_round_dot_setting
    @arg_validation_decos.is_line_round_dot_setting(arg_position_index=16)
    # line_dash_dot_setting
    @arg_validation_decos.is_line_dash_dot_setting(arg_position_index=17)
    # parent
    @arg_validation_decos.is_display_object_container(
        arg_position_index=18, optional=True
    )
    # variable_name_suffix
    @arg_validation_decos.is_builtin_string(arg_position_index=19, optional=False)
    @add_debug_info_setting(module_name=__name__)
    def __init__(
        self,
        *,
        x: Union[int, Int],
        y: Union[int, Int],
        width: Union[int, Int],
        height: Union[int, Int],
        ellipse_width: Union[int, Int] = 0,
        ellipse_height: Union[int, Int] = 0,
        fill_color: Union[str, String] = "",
        fill_alpha: Union[float, Number] = 1.0,
        line_color: Union[str, String] = "",
        line_alpha: Union[float, Number] = 1.0,
        line_thickness: Union[int, Int] = 1,
        line_cap: Optional[Union[String, LineCaps]] = None,
        line_joints: Optional[Union[String, LineJoints]] = None,
        line_dot_setting: Optional[LineDotSetting] = None,
        line_dash_setting: Optional[LineDashSetting] = None,
        line_round_dot_setting: Optional[LineRoundDotSetting] = None,
        line_dash_dot_setting: Optional[LineDashDotSetting] = None,
        parent: Optional[ChildMixIn] = None,
        variable_name_suffix: str = "",
    ) -> None:
        """
        Create a rectangle vector graphic.

        Parameters
        ----------
        x : Int or int
            X-coordinate to start drawing.
        y : Int or int
            Y-coordinate to start drawing.
        width : Int or int
            Rectangle width.
        height : Int or int
            Rectangle height.
        ellipse_width : int or Int
            Ellipse width.
        ellipse_height : int or Int
            Ellipse height.
        fill_color : str or String, default ''
            A fill-color to set.
        fill_alpha : float or Number, default 1.0
            A fill-alpha to set.
        line_color : str or String, default ''
            A line-color to set.
        line_alpha : float or Number, default 1.0
            A line-alpha to set.
        line_thickness : int or Int, default 1
            A line-thickness (line-width) to set.
        line_cap : String or LineCaps or None, default None
            A line-cap setting to set.
        line_joints : String or LineJoints or None, default None
            A line-joints setting to set.
        line_dot_setting : LineDotSetting or None, default None
            A dot setting to set.
        line_dash_setting : LineDashSetting or None, default None
            A dash setting to set.
        line_round_dot_setting : LineRoundDotSetting or None, default None
            A round-dot setting to set.
        line_dash_dot_setting : LineDashDotSetting or None, default None
            A dash dot (1-dot chain) setting to set.
        parent : ChildMixIn or None, default None
            A parent instance to add this instance.
            If a specified value is None, this interface uses
            a stage instance.
        variable_name_suffix : str, default ''
            A JavaScript variable name suffix string.
            This setting is sometimes useful for JavaScript debugging.

        References
        ----------
        - Rectangle class
            - https://simon-ritchie.github.io/apysc/en/rectangle.html

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> rectangle: ap.Rectangle = ap.Rectangle(
        ...     x=50, y=50, width=100, height=100, fill_color="#00aaff"
        ... )
        >>> rectangle.x
        Int(50)
        >>> rectangle.y
        Int(50)
        >>> rectangle.width
        Int(100)
        >>> rectangle.height
        Int(100)
        >>> rectangle.fill_color
        String('#00aaff')
        """
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names

        self._variable_name_suffix = variable_name_suffix
        variable_name: str = expression_variables_util.get_next_variable_name(
            type_name=var_names.RECTANGLE
        )
        self.variable_name = variable_name
        self._update_x_and_skip_appending_exp(x=x)
        self._update_y_and_skip_appending_exp(y=y)
        self._update_width_and_skip_appending_exp(value=width)
        self._update_height_and_skip_appending_exp(value=height)
        self._set_initial_basic_values(
            fill_color=fill_color,
            fill_alpha=fill_alpha,
            line_color=line_color,
            line_thickness=line_thickness,
            line_alpha=line_alpha,
            line_cap=line_cap,
            line_joints=line_joints,
        )
        self._append_constructor_expression()
        self._set_line_setting_if_not_none_value_exists(
            line_dot_setting=line_dot_setting,
            line_dash_setting=line_dash_setting,
            line_round_dot_setting=line_round_dot_setting,
            line_dash_dot_setting=line_dash_dot_setting,
        )
        self._set_ellipse_settings_if_values_are_not_zero(
            ellipse_width=ellipse_width, ellipse_height=ellipse_height
        )
        super(Rectangle, self).__init__(parent=parent, variable_name=variable_name)

    @final
    def _set_ellipse_settings_if_values_are_not_zero(
        self, *, ellipse_width: Union[int, Int], ellipse_height: Union[int, Int]
    ) -> None:
        """
        Set ellipse-related settings if values are not zero.

        Parameters
        ----------
        ellipse_width : Union[int, Int]
            Ellipse width to set.
        ellipse_height : Union[int, Int]
            Ellipse height to set.
        """
        if isinstance(ellipse_width, int) and ellipse_width == 0:
            return
        if isinstance(ellipse_height, int) and ellipse_height == 0:
            return
        if isinstance(ellipse_width, Int) and ellipse_width._value == 0:
            return
        if isinstance(ellipse_height, Int) and ellipse_height._value == 0:
            return
        if isinstance(ellipse_width, int):
            ellipse_width = Int(ellipse_width)
        if isinstance(ellipse_height, int):
            ellipse_height = Int(ellipse_height)
        self.ellipse_width = ellipse_width
        self.ellipse_height = ellipse_height

    @classmethod
    @final
    def _create_with_graphics(
        cls,
        *,
        graphics: "graphics.Graphics",
        x: Union[int, Int],
        y: Union[int, Int],
        width: Union[int, Int],
        height: Union[int, Int],
        variable_name_suffix: str = "",
    ) -> "Rectangle":
        """
        Create a rectangle instance with the instance of
        specified graphics..

        Parameters
        ----------
        graphics : Graphics
            Graphics instance to link this instance.
        x : Int or int
            X-coordinate to start drawing.
        y : Int or int
            Y-coordinate to start drawing.
        width : Int or int
            Rectangle width.
        height : Int or int
            Rectangle height.
        variable_name_suffix : str, default ''
            A JavaScript variable name suffix string.
            This setting is sometimes useful for JavaScript debugging.

        Returns
        -------
        rectangle : Rectangle
            A created rectangle instance.
        """
        rectangle: Rectangle = Rectangle(
            x=x,
            y=y,
            width=width,
            height=height,
            fill_color=graphics._fill_color,
            fill_alpha=graphics._fill_alpha,
            line_color=graphics._line_color,
            line_alpha=graphics._line_alpha,
            line_thickness=graphics._line_thickness,
            line_cap=graphics._line_cap,
            line_joints=graphics._line_joints,
            line_dot_setting=graphics._line_dot_setting,
            line_dash_setting=graphics._line_dash_setting,
            line_round_dot_setting=graphics._line_round_dot_setting,
            line_dash_dot_setting=graphics._line_dash_dot_setting,
            parent=graphics,
            variable_name_suffix=variable_name_suffix,
        )
        return rectangle

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Rectangle('<variable_name>')`).
        """
        repr_str: str = f"Rectangle('{self.variable_name}')"
        return repr_str

    @final
    @add_debug_info_setting(module_name=__name__)
    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap

        variable_name: str = self.variable_name
        stage: ap.Stage = ap.get_stage()
        expression: str = (
            f"var {variable_name} = {stage.variable_name}"
            f"\n  .rect({self.width.variable_name}, "
            f"{self.height.variable_name})"
            "\n  .attr({"
        )
        expression = self._append_basic_vals_expression(
            expression=expression, indent_num=2
        )
        expression += "\n  });"
        ap.append_js_expression(expression=expression)



"""Class implementation for the scale-x animation value.
"""

from typing import Generic
from typing import TypeVar
from typing import Union

from typing_extensions import final

from apysc._animation.animation_base import AnimationBase
from apysc._animation.easing import Easing
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.variable_name_mixin import VariableNameMixIn

_Target = TypeVar("_Target", bound=VariableNameMixIn)


class AnimationScaleXFromCenter(AnimationBase[_Target], Generic[_Target]):
    """
    The animation class for a scale-x.

    References
    ----------
    - animation_scale_x_from_center interface
        - https://simon-ritchie.github.io/apysc/en/animation_scale_x_and_y_from_center.html  # noqa
    - Animation interfaces duration setting
        - https://simon-ritchie.github.io/apysc/en/animation_duration.html
    - Animation interfaces delay setting
        - https://simon-ritchie.github.io/apysc/en/animation_delay.html
    - Each animation interface return value
        - https://simon-ritchie.github.io/apysc/en/animation_return_value.html  # noqa
    - Sequential animation setting
        - https://simon-ritchie.github.io/apysc/en/sequential_animation.html
    - animation_parallel interface
        - https://simon-ritchie.github.io/apysc/en/animation_parallel.html
    - Easing enum
        - https://simon-ritchie.github.io/apysc/en/easing_enum.html

    Examples
    --------
    >>> import apysc as ap
    >>> stage: ap.Stage = ap.Stage()
    >>> sprite: ap.Sprite = ap.Sprite()
    >>> sprite.graphics.begin_fill(color="#0af")
    >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
    ...     x=50, y=50, width=50, height=50
    ... )
    >>> animation: ap.AnimationScaleXFromCenter
    >>> animation = rectangle.animation_scale_x_from_center(
    ...     scale_x_from_center=0.5,
    ...     duration=1500,
    ...     easing=ap.Easing.EASE_OUT_QUINT,
    ... )
    >>> _ = animation.start()
    """

    _scale_x_from_center: Number
    _before_scale_x_from_center: Number
    _scale_x_from_center_diff_ratio: Number

    @final
    @add_debug_info_setting(module_name=__name__)
    def __init__(
        self,
        *,
        target: _Target,
        scale_x_from_center: Union[float, Number],
        duration: Union[int, Int] = 3000,
        delay: Union[int, Int] = 0,
        easing: Easing = Easing.LINEAR,
    ) -> None:
        """
        The animation class for a scale-x.

        Parameters
        ----------
        target : SkewXMixIn
            A target instance of the animation target
            (e.g., `Rectangle` instance).
        scale_x_from_center : float or Number
            The final scale-x of the animation.
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Raises
        ------
        TypeError
            If a specified target is not a ScaleXFromCenterMixIn
            instance.
        """
        import apysc as ap
        from apysc._converter import to_apysc_val_from_builtin
        from apysc._display.scale_x_from_center_mixin import ScaleXFromCenterMixIn
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names

        variable_name: str = expression_variables_util.get_next_variable_name(
            type_name=var_names.ANIMATION_SCALE_X_FROM_CENTER
        )
        target_: VariableNameMixIn = target
        if isinstance(target_, ScaleXFromCenterMixIn):
            target_._initialize_scale_x_from_center_if_not_initialized()
            self._before_scale_x_from_center = target_._scale_x_from_center
        else:
            raise TypeError(
                "Specified `target` argument is not a "
                f"ScaleXFromCenterMixIn instance: {type(target_)}"
            )
        self._scale_x_from_center = (
            to_apysc_val_from_builtin.get_copied_number_from_builtin_val(
                float_or_num=scale_x_from_center
            )
        )
        one: ap.Number = ap.Number(1.0)
        self._scale_x_from_center_diff_ratio = (
            one / self._before_scale_x_from_center * self._scale_x_from_center
        )
        self._set_basic_animation_settings(
            target=target, duration=duration, delay=delay, easing=easing
        )
        super(AnimationScaleXFromCenter, self).__init__(variable_name=variable_name)

    @final
    def _get_animation_func_expression(self) -> str:
        """
        Get a animation function expression.

        Returns
        -------
        expression : str
            Animation function expression.
        """
        from apysc._type import value_util

        diff_ratio_str: str = value_util.get_value_str_for_expression(
            value=self._scale_x_from_center_diff_ratio
        )
        return f"\n  .scale({diff_ratio_str}, 1);"

    @final
    def _get_complete_event_in_handler_head_expression(self) -> str:
        """
        Get an expression to be inserted into the complete event
        handler's head.

        Returns
        -------
        expression : str
            An expression to insert into the complete event
            handler's head.
        """
        from apysc._display.scale_x_from_center_mixin import ScaleXFromCenterMixIn

        expression: str = ""
        if isinstance(self._target, ScaleXFromCenterMixIn):
            self._target._initialize_scale_x_from_center_if_not_initialized()
            expression = (
                f"{self._target._scale_x_from_center.variable_name} = "
                f"{self._scale_x_from_center.variable_name};"
            )
        return expression



"""Class implementation for the polygon-related classes' x1 mix-in.
"""

from apysc._display.polygon_apply_current_points_mixin import (
    PolygonApplyCurrentPointsMixIn,
)
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.int import Int


class PolygonX1MixIn(PolygonApplyCurrentPointsMixIn):

    _x1: Int

    @property
    @add_debug_info_setting(module_name=__name__)
    def x1(self) -> Int:
        """
        Get a first x-coordinate.

        Returns
        -------
        x1 : Int
            A first x-coordinate.
        """
        x1: Int = self._points[0].x
        return x1

    @x1.setter
    @add_debug_info_setting(module_name=__name__)
    def x1(self, value: Int) -> None:
        """
        Set a first x-coordinate.

        Parameters
        ----------
        value : Int
            X-coordinate to set.
        """
        import apysc as ap

        self._x1._value = value._value
        new_point: ap.Point2D = self._points[0]
        new_point.x = value
        self._points[0] = new_point
        self._apply_current_points()



"""This module is for the translation mapping data of the
following document:

Document file: graphics_base_line_round_dot_setting.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {
    ##################################################
    "# GraphicsBase line_round_dot_setting interface": "# GraphicsBase クラスの line_round_dot_setting インターフェイス",  # noqa
    ##################################################
    "This page explains the `GraphicsBase` class `line_round_dot_setting` property interface.": "このページでは`GraphicsBase`クラスの`line_round_dot_setting`属性のインターフェイスについて説明します。",  # noqa
    ##################################################
    "## What interface is this?": "## インターフェイス概要",
    ##################################################
    "The `line_round_dot_setting` property interface updates or get the instance's current line round dot setting.": "`line_round_dot_setting`属性のインターフェイスではインスタンスの現在線のの丸ドットのスタイル設定の更新や取得を行えます。",  # noqa
    ##################################################
    "## Basic usage": "## 基本的な使い方",
    ##################################################
    "The getter or setter interface value becomes (or requires) the `LineRoundDotSetting` instance value.": "getterやsetterのインターフェイスの値は`interface`クラスのインスタンスの値になります。",  # noqa
    ##################################################
    "The following example sets the 10-pixel round size to the line:": "以下の例では10pxのサイズの丸ドットの線のスタイル設定を行っています:",  # noqa
    ##################################################
    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    stage_width=250, stage_height=100, background_color="#333", stage_elem_id="stage"\n)\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.line_style(color="#0af", thickness=5)\n\nline: ap.Line = sprite.graphics.draw_line(x_start=50, y_start=50, x_end=200, y_end=50)\nline.line_round_dot_setting = ap.LineRoundDotSetting(round_size=10, space_size=5)\n\nap.save_overall_html(\n    dest_dir_path="./graphics_base_line_round_dot_setting_basic_usage/"\n)\n```': '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    stage_width=250, stage_height=100, background_color="#333", stage_elem_id="stage"\n)\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.line_style(color="#0af", thickness=5)\n\nline: ap.Line = sprite.graphics.draw_line(x_start=50, y_start=50, x_end=200, y_end=50)\nline.line_round_dot_setting = ap.LineRoundDotSetting(round_size=10, space_size=5)\n\nap.save_overall_html(\n    dest_dir_path="./graphics_base_line_round_dot_setting_basic_usage/"\n)\n```',  # noqa
    ##################################################
    "## line_round_dot_setting property API": "## line_round_dot_setting 属性のAPI",
    ##################################################
    '<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>': '<span class="inconspicuous-txt">特記事項: このAPIドキュメントはドキュメントビルド用のスクリプトによって自動で生成・同期されています。そのためもしかしたらこの節の内容は前節までの内容と重複している場合があります。</span>',  # noqa
    ##################################################
    "**[Interface summary]**": "**[インターフェイス概要]**",
    ##################################################
    "Get this instance's line round dot setting.<hr>": "**[インターフェイス]** インスタンスの線の丸ドットのスタイル設定を取得します。<hr>",  # noqa
    ##################################################
    "**[Returns]**": "**[返却値]**",
    ##################################################
    "- `line_round_dot_setting`: LineRoundDotSetting or None": "- `line_round_dot_setting`: LineRoundDotSetting or None",  # noqa
    ##################################################
    "  - Line round dot setting.": "  - 線の丸ドットのスタイル設定。",
    ##################################################
    "<hr>": "<hr>",
    ##################################################
    "**[Examples]**": "**[コードサンプル]**",
    ##################################################
    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.line_style(color="#fff", thickness=10)\n>>> line: ap.Line = sprite.graphics.draw_line(\n...     x_start=50, y_start=50, x_end=150, y_end=50\n... )\n>>> line.line_round_dot_setting = ap.LineRoundDotSetting(\n...     round_size=10, space_size=5\n... )\n>>> line.line_round_dot_setting.round_size\nInt(10)\n\n>>> line.line_round_dot_setting.space_size\nInt(5)\n```': '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.line_style(color="#fff", thickness=10)\n>>> line: ap.Line = sprite.graphics.draw_line(\n...     x_start=50, y_start=50, x_end=150, y_end=50\n... )\n>>> line.line_round_dot_setting = ap.LineRoundDotSetting(\n...     round_size=10, space_size=5\n... )\n>>> line.line_round_dot_setting.round_size\nInt(10)\n\n>>> line.line_round_dot_setting.space_size\nInt(5)\n```',  # noqa
}



"""Class implementations for the month-related mix-in.
"""

from typing import Union

from typing_extensions import final

from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.int import Int
from apysc._type.revert_mixin import RevertMixIn
from apysc._type.variable_name_mixin import VariableNameMixIn
from apysc._type.variable_name_suffix_attr_or_var_mixin import (
    VariableNameSuffixAttrOrVarMixIn,
)
from apysc._validation import arg_validation_decos


class MonthMixIn(VariableNameMixIn, VariableNameSuffixAttrOrVarMixIn, RevertMixIn):

    _initial_month: Union[int, Int]
    _month: Int

    @final
    @add_debug_info_setting(module_name=__name__)
    @arg_validation_decos.is_month_int(arg_position_index=1)
    def _set_init_month_value(self, *, month: Union[int, Int]) -> None:
        """
        Set an initial month value.

        Parameters
        ----------
        month : Union[int, Int]
            A month value to set.
        """
        from apysc._converter.to_apysc_val_from_builtin import (
            get_copied_int_from_builtin_val,
        )

        self._initial_month = month
        suffix: str = self._get_attr_or_variable_name_suffix(value_identifier="month")
        self._month = get_copied_int_from_builtin_val(
            integer=month, variable_name_suffix=suffix
        )

    @final
    @add_debug_info_setting(module_name=__name__)
    def _get_init_month_argument_expression(self) -> str:
        """
        Get an initial month's argument expression string.

        Returns
        -------
        expression : str
            A created expression string.
        """
        if isinstance(self._initial_month, Int):
            return f", {self._initial_month.variable_name} - 1"

        return f", {self._month._value - 1}"

    @property
    @add_debug_info_setting(module_name=__name__)
    def month(self) -> Int:
        """
        Get a current month's value.

        Returns
        -------
        month : Int
            A current month value.

        References
        ----------
        - DateTime class month property
            - https://simon-ritchie.github.io/apysc/en/datetime_month.html

        Examples
        --------
        >>> import apysc as ap
        >>> datetime_: ap.DateTime = ap.DateTime(year=2022, month=12, day=1)
        >>> datetime_.month
        Int(12)
        >>> datetime_.month = ap.Int(1)
        >>> datetime_.month
        Int(1)
        """
        copied_month_val: Int = self._month._copy()
        self._append_month_getter_expression(month_val=copied_month_val)
        return copied_month_val

    @month.setter
    @arg_validation_decos.is_month_int(arg_position_index=1)
    def month(self, value: Int) -> None:
        """
        Set a month value.

        Parameters
        ----------
        value : Int
            A month value to set.

        References
        ----------
        - DateTime class month property
            - https://simon-ritchie.github.io/apysc/en/datetime_month.html

        Examples
        --------
        >>> import apysc as ap
        >>> datetime_: ap.DateTime = ap.DateTime(year=2022, month=12, day=1)
        >>> datetime_.month
        Int(12)
        >>> datetime_.month = ap.Int(1)
        >>> datetime_.month
        Int(1)
        """
        self._month = value._copy()
        self._append_month_setter_expression(month_val=value)

    @final
    @arg_validation_decos.is_month_int(arg_position_index=1)
    @add_debug_info_setting(module_name=__name__)
    def _append_month_getter_expression(self, *, month_val: Int) -> None:
        """
        Append a month's getter expression string.

        Parameters
        ----------
        month_val : Int
            A month value to use in an expression.
        """
        import apysc as ap

        expression: str = (
            f"{month_val.variable_name} = {self.variable_name}.getMonth() + 1;"
        )
        ap.append_js_expression(expression=expression)

    @final
    @add_debug_info_setting(module_name=__name__)
    def _append_month_setter_expression(self, *, month_val: Int) -> None:
        """
        Append a month's setter expression string.

        Parameters
        ----------
        month_val : Int
            A month value to use in an expression.
        """
        import apysc as ap

        expression: str = (
            f"{self.variable_name}.setMonth({month_val.variable_name} - 1);"
        )
        ap.append_js_expression(expression=expression)

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._month._run_all_make_snapshot_methods(snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if a snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._month._run_all_revert_methods(snapshot_name=snapshot_name)
